/* exons2underlays.c -- convert an exons file to underlay format
 *
 * This program was suggested by Mike Wilson.
 *
 * $Revision: 1.23 $$Date: 2001/04/30 18:53:08 $
 */

#include "exons.h"
#include "util.h"

#define SYNTAX  "exons_file [-laj] [-split] " \
                  "-fexon fwd_exon_color [-futr fwd_utr_color] " \
                  "[-rexon rev_exon_color] [-rutr rev_utr_color] " \
                  "[-intron intron_color] > output_file"

#define DEFAULT_INTRON_COLOR  "White"

static char *Types[5] = {
	"fwd_exon", "fwd_UTR", "rev_exon", "rev_UTR", "intron"
};

static void get_args (int argc, char **argv, char **filename, bool *laj,
                      bool *split, char **colors);
static void print_underlay (int begin, int end, char *name,
                            char *exon_name, int exon_num, int dir,
                            bool laj, bool split);

int main (int argc, char **argv) {
	char *filename = NULL;                  /* name of exons file           */
	bool laj = 0;                           /* make labeled underlays?      */
	bool split = 0;                         /* split strands?               */
	char *colors[5] = {NULL, NULL, NULL,    /* color names, in SYNTAX order */
	                   NULL, NULL};
	char *header;                           /* header line from exons file  */
	char **gene_name, **exon_name;          /* arrays of gene/exon info     */
	int *begin, *end, *dir, *exon_num;      /* arrays of gene/exon info     */
	int length;                             /* length of arrays             */
	int i, j;                               /* array indices                */
	char *gname;                            /* name of current gene         */
	int cds_begin, cds_end;                 /* translated region            */
	int num;                                /* exon number within gene      */

	/* get arguments */
	set_argv0 (argv[0]);
	get_args (argc, argv, &filename, &laj, &split, colors);
	if (!filename)
		fatal (usage (SYNTAX));
	if (!colors[0] && !colors[2])     /* exons */
		fatal (usage (SYNTAX));
	else if (!colors[0])
		colors[0] = colors[2];
	else if (!colors[2])
		colors[2] = colors[0];
	if (!colors[1] && !colors[3]) {   /* utrs */
		colors[1] = colors[0];
		colors[3] = colors[2];
	}
	else if (!colors[1])
		colors[1] = colors[3];
	else if (!colors[3])
		colors[3] = colors[1];
	if (!colors[4])                   /* introns */
		colors[4] = DEFAULT_INTRON_COLOR;

	/* print header */
	header = exon_header (filename);
	printf ("# Underlays converted from \"%s\" by %s.\n", filename, argv0);
	printf ("# %s\n", (header ? header : "(no header specified)"));

	/* print color legend */
	putchar ('\n');
	for (i = 0; i < 5; i++) {
		assert (colors[i]);
		printf ("%s %s\n", colors[i], Types[i]);
	}

	/* read gene and exon info */
	length = get_exons (filename, &begin, &end, &dir, &gene_name,
	                    &exon_name, -1);    /* print all errors */

	/* compute exon numbers */
	exon_num = (int *) ckalloc (length * sizeof (int));
	assert (!length || gene_name[0]);
	for (i = 0; i < length; ) {
		num = 0;
		if (dir[i] < 0) {
			for (j = i+2; j < length && !gene_name[j]; j++)
				num++;
		}
		for (j = i+2; j < length && !gene_name[j]; j++)
			exon_num[j] = (dir[j] < 0 ? num-- : ++num);
		i = j;
	}

	/* print underlays */
	gname = NULL; cds_begin = cds_end = 0;    /* pacify lint */
	for (i = 0; i < length; i++) {
		if (gene_name[i]) {     /* gene */
			gname = gene_name[i];
			printf ("\n");
			print_underlay (begin[i], end[i], gname, NULL, 0, dir[i],
			                laj, split);
			assert (++i < length);     /* cds */
			cds_begin = begin[i];
			cds_end = end[i];
		}
		else {     /* exon */
			print_underlay (begin[i], end[i], gname, exon_name[i],
			                exon_num[i], dir[i], laj, split);
			if (begin[i] < cds_begin)
				print_underlay (begin[i], MIN (cds_begin-1, end[i]),
				                (dir[i] < 0 ? "3'UTR" : "5'UTR"),
				                NULL, 0, dir[i], laj, split);
			if (cds_end > 0 && cds_end < end[i])
				print_underlay (MAX (cds_end+1, begin[i]), end[i],
				                (dir[i] < 0 ? "5'UTR" : "3'UTR"),
				                NULL, 0, dir[i], laj, split);
		}

	}

	/* clean up */
	free_exons (begin, end, dir, gene_name, exon_name, length);
	free (exon_num);
	return 0;
}

static void get_args (int argc, char **argv, char **filename, bool *laj,
                      bool *split, char **colors) {
	int i;

	for (i = 1; i < argc; i++) {
		if (strsame (argv[i], "-laj"))
			*laj = 1;
		else if (strsame (argv[i], "-split"))
			*split = 1;
		else if (strsame (argv[i], "-fexon"))
			colors[0] = optval (argc, argv, &i, SYNTAX);
		else if (strsame (argv[i], "-futr"))
			colors[1] = optval (argc, argv, &i, SYNTAX);
		else if (strsame (argv[i], "-rexon"))
			colors[2] = optval (argc, argv, &i, SYNTAX);
		else if (strsame (argv[i], "-rutr"))
			colors[3] = optval (argc, argv, &i, SYNTAX);
		else if (strsame (argv[i], "-intron"))
			colors[4] = optval (argc, argv, &i, SYNTAX);
		else if (!*filename)
			*filename = argv[i];
		else
			fatal (usage (SYNTAX));
	}
}

static void print_underlay (int begin, int end, char *name,
                            char *exon_name, int exon_num, int dir,
                            bool laj, bool split) {
	int type;
	char *half;

	if (strstr (name, "'UTR"))
		type = (dir < 0 ? 3 : 1);
	else if (exon_num > 0)
		type = (dir < 0 ? 2 : 0);
	else
		type = 4;

	half = (split && dir ? (dir < 0 ? "-" : "+") : "");

	if (laj) {
		if (exon_name)
			printf ("%d %d (%s) %s dummy %s\n",
			        begin, end, exon_name, Types[type], half);
		else if (exon_num > 0)
			printf ("%d %d (exon%d) %s dummy %s\n",
			        begin, end, exon_num, Types[type], half);
		else
			printf ("%d %d (%s) %s dummy %s\n",
			        begin, end, name, Types[type], half);
	}
	else
		printf ("%d %d %s %s\n", begin, end, Types[type], half);
}
