/* find-cpg.c -- make "repeats"-style entries for CpG islands
 *
 * The PipMaker server automatically locates CpG islands in a
 * submitted sequence, but Laj does not.  If you want Laj to
 * display CpG islands the way PipMaker would, you can run this
 * program to find them, and attach its output to the end of
 * your repeats file.  (The repeats file describes all of the
 * non-exon feature symbols that appear in the panel above the
 * pip, even if they are not actually repeats.)
 *
 * This program uses criteria that match PipMaker's (at least,
 * as of Dec. 2000).  That is, it finds regions of at least 50%
 * G+C content where the ratio of CpG dinucleotides relative to
 * GpC is at least 75 or 60 percent.  Like PipMaker, Laj will
 * draw these as low gray and white boxes, respectively.
 * 
 * $Revision: 1.9 $$Date: 2001/04/04 21:39:49 $
 */

#include "seq.h"
#include "util.h"

#define SYNTAX  "seq_file > output_file"
#define TAG_1   "CpG60"
#define TAG_2   "CpG75"

enum { MIN_GC = 50, RATIO_1 = 60, RATIO_2 = 75, WINDOW = 200, SHIFT = 100 };

static void get_args (int argc, char **argv, char **seqfile);
static void find_cpg (uchar *seq, int seqlen, int min_ratio, char *tag);
static void record (int start, int end, int *tot_start, int *tot_end,
                    char *tag);

int main (int argc, char **argv) {
	char *seqfile = NULL;                   /* name of input file           */
	SEQ *sp;                                /* ptr to struct holding seq    */
	uchar *seq;                             /* sequence characters          */
	int seqlen;                             /* length of sequence           */

	/* get arguments */
	set_argv0 (argv[0]);
	get_args (argc, argv, &seqfile);
	if (!seqfile)
		fatal (usage (SYNTAX));

	/* read sequence */
	if (!(sp = seq_get (seqfile)))
		fatalf ("Couldn't get sequence \"%s\".", seqfile);
	seq = SEQ_CHARS (sp);
	seqlen = SEQ_LEN (sp);
	if (SEQ_FROM (sp) != 1)
		fatal ("Specification of subsequence is not supported.");

	/* find & print islands */
	find_cpg (seq, seqlen, RATIO_1, TAG_1);
	find_cpg (seq, seqlen, RATIO_2, TAG_2);

	/* warn if additional (unprocessed) sequences */
	if (seq_read (sp) > 0) {
		fprintf (stderr, "Warning: %s does not support multiple contigs\n"
		         "in the sequence file -- only the first was used.\n", argv0);
	}

	/* clean up */
	seq_close (sp);
	return 0;
}

/* retrieve command-line arguments */
static void get_args (int argc, char **argv, char **seqfile) {
	int i;

	for (i = 1; i < argc; i++) {
		if (!*seqfile)
			*seqfile = argv[i];
		else
			fatal (usage (SYNTAX));
	}
}

/* find islands meeting minimum G+C content and CpG/GpC ratio */
static void find_cpg (uchar *seq, int seqlen, int min_ratio, char *tag) {
	int start, end, len, tot_start, tot_end;
	int i, nGC, nCpG, nGpC, ratio;
	uchar c, cc;

	tot_start = tot_end = -1;   /* cumulative interval set to empty */
	for (start = end = 0; end < seqlen; start += SHIFT) {
		end = start + WINDOW;
		if (end + SHIFT >= seqlen)
			end = seqlen;
		len = end - start;
		nGC = nCpG = nGpC = 0;
		for (i = start; i < end; i++) {
			c = seq[i];
			cc = (i > 0 ? seq[i-1] : '\0');
			if (c == 'G' || c == 'C')
				nGC++;
			if (cc == 'C' && c == 'G')
				nCpG++;
			if (cc == 'G' && c == 'C')
				nGpC++;
		}
		ratio = (nGpC > 0 ? (nCpG*100)/nGpC : nCpG*100);
		if ((nGC*100)/len >= MIN_GC && ratio >= min_ratio)
			record (start, end, &tot_start, &tot_end, tag);
	}
	/* flush the last one */
	record (seqlen+1, seqlen+1, &tot_start, &tot_end, tag);
}

/* write to output, merging overlapping islands */
static void record (int start, int end, int *tot_start, int *tot_end,
                    char *tag) {
	if (*tot_start < 0) {     /* first call */
		*tot_start = start;
	}
	else if (start > *tot_end) {
		printf ("%d %d %s\n", *tot_start, *tot_end, tag);
		*tot_start = start;
	}
	*tot_end = end;
}

